#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <termios.h>

pid_t child_pid;

void signal_handler(int number)
{
    printf("Caught signal %d, propagating to the child\n", number);
    if (kill(child_pid, number) == -1) {
        perror("kill failed");
    }
}

int main(int argc, char* argv[])
{
    if (argc < 2) {
        // there must be at least a single argument
        fprintf(stderr, "no executable specified\n");
        return EXIT_FAILURE;
    }
    
    struct timespec start;
    if (clock_gettime(CLOCK_MONOTONIC_RAW, &start)) {
        perror("clock_gettime failed on start");
    }
    
    child_pid = fork();
    if (child_pid == -1) {
        // fork failed
        perror("fork failed");
        return EXIT_FAILURE;
    }
    else if (child_pid == 0) {
        // executing child; replace the executable
        execv(argv[1], argv + 1);
        perror("execv failed");
        return EXIT_FAILURE;
    }
    // parent, install the signal handlers
    struct sigaction sa;
    sa.sa_handler = signal_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    if (sigaction(SIGINT, &sa, NULL) == -1) {
        perror("installing SIGINT handler failed");
    }
    if (sigaction(SIGTERM, &sa, NULL) == -1) {
        perror("installing SIGTERM handler failed");
    }
    
    // wait for the child and report
    int chld_state;
    pid_t rc_pid;
    do {
        rc_pid = waitpid(child_pid, &chld_state, 0);
        if (rc_pid == -1 && errno != EINTR) {
            perror("waitpid failed");
            return EXIT_FAILURE;
        }
    } while (rc_pid == -1);
    
    // capture the end time
    struct timespec end;
    if (clock_gettime(CLOCK_MONOTONIC_RAW, &end)) {
        perror("clock_gettime failed at the end");
    }
    
    // print the exit status
    printf("================================\n");
    if (WIFEXITED(chld_state)) {
        printf("Program exited with status %d\n", WEXITSTATUS(chld_state));
        
        // print execution time
        long long seconds, milliseconds;
        if (end.tv_nsec > start.tv_nsec) {
            seconds = end.tv_sec - start.tv_sec;
            milliseconds = (end.tv_nsec - start.tv_nsec)/1000000;
        }
        else {
            seconds = end.tv_sec - start.tv_sec - 1;
            milliseconds = (end.tv_nsec - start.tv_nsec)/1000000 + 1000;
        }
        printf("Execution took %lld.%03llds\n", seconds, milliseconds);
    }
    else if (WIFSIGNALED(chld_state)) {
        printf("Program terminated with signal %d: %s\n", WTERMSIG(chld_state), strsignal(WTERMSIG(chld_state)));
    }
    else {
        fprintf(stderr, "Something strange happened.\n");
        return EXIT_FAILURE;
    }

    // wait for a keypress
    printf("\nPress any key to continue...\n");
    struct termios info;
    tcgetattr(0, &info);
    info.c_lflag &= ~ICANON;
    info.c_cc[VMIN] = 1;
    info.c_cc[VTIME] = 0;
    tcsetattr(0, TCSANOW, &info);
    
    getchar();
    
    return EXIT_SUCCESS;
}
